#!/usr/bin/python

import getopt
import math
import os
import re
import sys

import matplotlib.font_manager
import matplotlib.pyplot
import numpy

FONT_FAMILY = 'Bitstream Vera Sans Serif'
FONT_SIZE = 7
FONT_LEGEND = matplotlib.font_manager.FontProperties(FONT_FAMILY, size=6)
DIMEN = (8, 5)

SCALE_LINEAR = 0
SCALE_LOG = 1 # needs 'ymin' to be set

GRAPH_TASK_SYS = 0
GRAPH_TASK_LANG = 1
GRAPH_FULL_PLATFORM = 2
GRAPHPARAM = {
	GRAPH_TASK_SYS: {
		'short': 'task_sys',
		'yscale': SCALE_LINEAR,
		'xlabel': 'System',
		'ylabel': 'Time (s)',
		'data': [
			('App Engine', 1.69),
			('Xeon E5520 (2.26 GHz)', 1.37),
			('Phenom II 955 (3.2 GHz)', 1.18),
			('Atom N270', 9.85),
		],
	}, GRAPH_TASK_LANG: {
		'short': 'task_lang',
		'yscale': SCALE_LINEAR,
		'xlabel': 'Language',
		'ylabel': 'Time (s)',
		'data': [
			('C++', 0.0126),
			('Python', 1.18),
		],
	}, GRAPH_FULL_PLATFORM: {
		'short': 'full_platform',
		'yscale': SCALE_LOG,
		'xlabel': 'Platform',
		'ylabel': 'Time (s)',
		'ymin': 10,
		'data': [
			('C++', 20.122),
			('Python (App Engine)', 275),
			('Python (local)', 2141.332),
		],
	},
}

# inverse
GRAPHTYPE = {}
for type in GRAPHPARAM.keys():
	GRAPHTYPE[GRAPHPARAM[type]['short']] = type

def plot(graphtype, formats, outdir, location='ur'):
	graphparam = GRAPHPARAM[graphtype]

	titles = [ l for (l,_) in graphparam['data'] ]
	data = [ d for (_,d) in graphparam['data'] ]

	color_darkest = 1.0 / (len(data) * 2 + 1)
	color_lightest = 1 - color_darkest
	colors = [ (x,x,x) for x in numpy.linspace(
	    color_lightest, color_darkest, num=len(data)) ]

	# the width, spacing of the bars
	width = 1. / (len(graphparam['data']) + 1)
	space = width / len(graphparam['data'])
	left_offsets = [ space / 2 + (width + space) * i for i in
	    range(len(data)) ]

	fig = matplotlib.pyplot.figure()
	fig.set_size_inches(DIMEN[0] / 2.54, DIMEN[1] / 2.54)

	axes = fig.add_subplot(111)

	rect_sets = []
	log = graphparam['yscale'] == SCALE_LOG
	for (val, color, offset) in zip(data, colors, left_offsets):
		rect_sets.append(axes.bar(offset, val, width,
		    color=color, log=log))

	if graphparam['yscale'] == SCALE_LINEAR:
		axes.set_yscale('linear')
	elif graphparam['yscale'] == SCALE_LOG:
		axes.set_ylim(ymin=graphparam['ymin'])
	else:
		assert 0

	axes.set_ylabel(graphparam['ylabel'], family=FONT_FAMILY,
	    size=FONT_SIZE)
	axes.set_xlim(0, 1)
	axes.set_xticks([])
	axes.set_xlabel(graphparam['xlabel'],
	    family=FONT_FAMILY, size=FONT_SIZE)

	for label in axes.get_yticklabels():
		label.set_family(FONT_FAMILY)
		label.set_size(FONT_SIZE)

	if location == 'ul':
		loc = 'upper left'
	elif location == 'ur':
		loc = 'upper right'
	elif location == 'll':
		loc = 'lower left'
	elif location == 'lr':
		loc = 'lower right'
	else:
		loc = None

	if loc:
		axes.legend([r[0] for r in rect_sets], titles,
		    loc=loc, prop=FONT_LEGEND)

	for ext in formats:
		fig.savefig(os.path.join(outdir, '%s.%s' % (
		    graphparam['short'], ext)))

prog = sys.argv[0]
for task in (
    GRAPH_TASK_SYS,
#    GRAPH_TASK_LANG,
    GRAPH_FULL_PLATFORM,
    ):
	plot(task, ['eps'], '.', location='ul')
