#!/usr/bin/python

import getopt
import math
import os
import re
import sys

import matplotlib.font_manager
import matplotlib.pyplot

FONT_FAMILY = 'Bitstream Vera Sans Serif'
FONT_LEGEND = matplotlib.font_manager.FontProperties(FONT_FAMILY, size=6)
FONT_SIZE = 7 # for labels
DIMEN = (16.5, 7) # limit seems to be ~16.54cm for letterpaper

LINE_REQ = 1
LINE_CPU = 2
LINE_TOT = 4
LINEPARAM = {
	LINE_CPU: {
		'format': '-',
		'color': '#000000',
		'label': 'CPU time',
		'short': 'cpu'
	},
	LINE_TOT: {
		'format': '-',
		'color': '#808080',
		'label': 'Total execution time',
		'short': 'tot'
	},
	LINE_REQ: {
		'format': '-',
		'color': '#c0c0c0',
		'label': 'Request delay',
		'short': 'req'
	},
}

SCALE_LIN = 0
SCALE_LOG2 = 1
SCALE_LOG10 = 2

GRAPH_REQ_TIMING = 0
GRAPHPARAM = {
	GRAPH_REQ_TIMING: {
		'range': (None, None),
		'scale': SCALE_LIN,
		'short': 'req_timing',
		'xlabel': 'Task number',
		'ylabel': 'Time (s)',
	},
}

def read_timing_file(path):
	f = open(path, 'r')
	exp = re.compile(
	    # 1 to 3 tab-delimited values
	    r'^ ([^ \t]+) (?: [ \t]+ ([^ \t]+) (?: [ \t]+ ([^ \t]+) )?)? $',
	    re.X)
	first = True
	req = []
	cpu = []
	total = []
	n = 0
	while True:
		line = f.readline()
		if not len(line): break
		if first:
			# skip first line
			first = False
			continue
		m = exp.match(line)
		if not m:
			print 'could not match `%s\'' % line
			continue

		req.append((n, float(m.group(1))))
		if m.group(2) is not None:
			cpu.append((n, float(m.group(2))))
		if m.group(3) is not None:
			total.append((n, float(m.group(3))))

		n += 1
	return (req, cpu, total)

def std_dev(vals):
	mean = sum(vals) / float(len(vals))
	sq_sum = sum( x*x for x in vals )
	return math.sqrt(sq_sum / float(len(vals)) - mean * mean)

def plot(graphtype, path, formats, outdir):
	graphparam = GRAPHPARAM[graphtype]

	# depending on the vartype, there may be multiple graphs; this is
	# realized in the call to data_paths()
	graphs = []
	plots = []

	data = read_timing_file(path)

	fig = matplotlib.pyplot.figure()
	fig.set_size_inches(DIMEN[0] / 2.54,
	    DIMEN[1] / 2.54)

	for line in (
	    #LINE_REQ,
	    LINE_CPU,
	    LINE_TOT,
	    ):
		lineparam = LINEPARAM[line]

		if line == LINE_REQ:
			data_idx = 0
		elif line == LINE_CPU:
			data_idx = 1
		elif line == LINE_TOT:
			data_idx = 2
		else:
			assert 0

		axes = fig.add_subplot(111)
		xdata = [ x[0] for x in data[data_idx] ]
		ydata = [ x[1] for x in data[data_idx] ]
		plots.append(axes.plot(xdata, ydata, lineparam['format'],
		    color=lineparam['color'],
		    label=lineparam['label']))

	# y-axis parameters
	axes.set_ylabel(graphparam['ylabel'],
	    family=FONT_FAMILY, size=FONT_SIZE)
	range_ = graphparam['range']
	axes.set_ybound(range_[0], range_[1])
	scale = graphparam['scale']
	if scale == SCALE_LOG2:
		axes.set_yscale('log', basey=2)
	elif scale == SCALE_LOG10:
		axes.set_yscale('log', basey=10)

	# x-axis parameters
	axes.set_xlabel('Request number',
	    family=FONT_FAMILY, size=FONT_SIZE)

	# set fonts of all tick labels; rescaling no longer possible
	for labels in (
	    axes.get_yticklabels(), axes.get_xticklabels()):
		for label in labels:
			label.set_family(FONT_FAMILY)
			label.set_size(FONT_SIZE)

	axes.legend(prop=FONT_LEGEND, loc='upper left')

	for ext in formats:
		fig.savefig(os.path.join(outdir,
		    '%s.%s' % (graphparam['short'], ext)))

prog = sys.argv[0]
plot(GRAPH_REQ_TIMING, 'timing.dat', ['eps'], '.')
