# -*- coding: utf-8 -*-

# Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
# Lena Olson <lena@cs.wisc.edu>
#
# You can't do anything with this but read and execute it privately.  If you
# want to do more, ask us if we'd be willing to make a more permissive
# license.

from array import array
import logging
import struct
import sys
import time
import zlib

from google.appengine.api.labs import taskqueue
from google.appengine.ext import db
from google.appengine.ext import webapp
from google.appengine.ext.webapp import template
from google.appengine.ext.webapp.util import run_wsgi_app
from google.appengine.runtime import DeadlineExceededError, apiproxy_errors

import datastore
import util

# same as in master.py
ASCII_MIN = 0x20
ASCII_MAX = 0x7f # non-inclusive
ASCII_RANGE = ASCII_MAX - ASCII_MIN # number of possible values

def incrByOneAscii(buf):
	'''Increment by one in place.'''
	pos = len(buf) - 1
	# ripple-carry :)
	while pos >= 0:
		sum = buf[pos] + 1
		if sum == ASCII_MAX:
			buf[pos] = ASCII_MIN
		else:
			buf[pos] = sum
			return
		pos -= 1
	buf.insert(0, ASCII_MIN)

def incrByOneBinary(buf):
	'''Increment by one in place.'''
	pos = len(buf) - 1
	# ripple-carry :)
	while pos >= 0:
		if buf[pos] == 0xff:
			buf[pos] = 0
		else:
			buf[pos] += 1
			return
		pos -= 1
	buf.insert(0, 0)

def task_loop(tgt_crc, mod_crc, utgt_crc, begin, end_base64, incrByOne):
	end = util.decode_array(end_base64)
	i = 0
	while begin != end:
		i += 1
		if i == 16384:
			i = 0
			if util.check_cache(utgt_crc): break

			util.set_task_progress(utgt_crc, end_base64,
			    util.encode_array(begin), True)

		# take the crc
		my_crc = zlib.crc32(begin, mod_crc)

		# check; if the same, update datastore; if not, continue
		if tgt_crc == my_crc:
			logging.info('FOUND IT %s' % util.encode_array(begin))

			# store this suffix (datastore knows it is done)
			datastore.add_suffix(utgt_crc,
			    util.encode_array(begin))

			# update the memcache to indicate that it is done
			util.add_completed(utgt_crc)

			break
		incrByOne(begin)

class collisionWorker(webapp.RequestHandler):
	def post(self): # should run at most 1/s
		time_start = time.time()

		utgt_crc = int(self.request.get('crc'))
		umod_crc = int(self.request.get('mod_crc'))
		if util.check_cache(utgt_crc): return
 
		#print >>sys.stderr, >>sys.stderr, (
		#    self.request.get('begin'),
		#    self.request.get('end'))

		begin = util.decode_array(self.request.get('begin'))
		end_base64 = self.request.get('end')
		append_mode = self.request.get('append_mode')

		logging.info('(begin,end)=(%s,%s)' %
		    (self.request.get('begin'), end_base64))

		tgt_crc = utgt_crc
		mod_crc = umod_crc
		# TODO the following logic needs to change so that 'begin' and
		# 'end' are of arbitrary length

		# if 32bit arch, zlib.crc32() may return negative numbers
		if type(0x80000000) == long:
			if tgt_crc & 0x80000000: tgt_crc -= 0x100000000
			if mod_crc & 0x80000000: mod_crc -= 0x100000000

		try:
			if append_mode == 'ascii':
				incr = incrByOneAscii
			elif append_mode == 'binary':
				incr = incrByOneBinary
			else: assert 0
			task_loop(tgt_crc, mod_crc, utgt_crc,
			    begin, end_base64, incr)

			# remove task from datastore
			datastore.remove_task(utgt_crc, end_base64)

			time_total = time.time() - time_start
			logging.info('TOTAL TIME %f' % time_total)

			return
		except DeadlineExceededError: pass
		except apiproxy_errors.DeadlineExceededError: pass

		# save progress of task
		logging.info('emergency begin update %s => %s' % (
		    self.request.get('begin'), util.encode_array(begin)))
		util.set_task_progress(utgt_crc, end_base64,
		    util.encode_array(begin), False)

		time_total = time.time() - time_start
		logging.info('TOTAL TIME %f' % time_total)

def main():
	run_wsgi_app(webapp.WSGIApplication([
	('/tryToCollide', collisionWorker)]))

if __name__ == '__main__':
	main()
