# -*- coding: utf-8 -*-

# Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
# Lena Olson <lena@cs.wisc.edu>
#
# You can't do anything with this but read and execute it privately.  If you
# want to do more, ask us if we'd be willing to make a more permissive
# license.

import base64
import logging
import sys
import time

from google.appengine.ext import blobstore
from google.appengine.ext import db

class Job(db.Model):
	crc = db.IntegerProperty(required=True)
	mod_crc = db.IntegerProperty(indexed=False, required=True)

	# one of these should be present
	blob_info = blobstore.BlobReferenceProperty(indexed=False)
	data = db.BlobProperty(indexed=False)

	append_mode = db.CategoryProperty(indexed=False,required=True)

	# reset to int(time.time()) when (1) the Master times out or (2) the
	# Master discovers a collision
	timestamp = db.IntegerProperty(required=True, indexed=False)

	# last 'end' value for any Task added, base64-encoded
	last_end = db.ByteStringProperty(required=True, indexed=False)

	# undefined at start, base64-encoded
	suffix = db.ByteStringProperty(indexed=False)

class Task(db.Model):
	crc = db.IntegerProperty(required=True)
	mod_crc = db.IntegerProperty(indexed=False, required=True)

	# begin, end: base64-encoded
	begin = db.ByteStringProperty(indexed=False, required=True)
	end = db.ByteStringProperty(required=True)

class Word(db.Model):
	word = db.StringProperty(required=True)

# JOB FUNCTIONS

def add_job(crc, mod_crc, append_mode, data=None, blob_info=None):
	'''Add to the Job table.  Either 'data' or 'blob_info' are
	required.'''
	assert data or blob_info

	# for testing 'alpha', 'beta'
	#last_end = base64.b64encode('7K%  ')
	last_end = base64.b64encode(' ')

	ent = Job(crc=crc, mod_crc=mod_crc, data=data, blob_info=blob_info,
	    append_mode=append_mode,
	    last_end=last_end, timestamp=int(time.time()))
	ent.put()

def get_job(crc):
	query = db.GqlQuery('SELECT * FROM Job WHERE crc = :1', crc)
	return query.get()

def job_complete(crc):
	'''Do a minimal check to see if a job is done.'''

	query = db.GqlQuery('''
	    SELECT * FROM Job
	    WHERE crc = :1''', crc)
	ent = query.get()
	return ent is not None and ent.suffix is not None

def update_timestamp(crc, timestamp, last_end):
	def tx(key, timestamp, last_end):
		job = db.get(key)
		job.timestamp = timestamp
		job.last_end = last_end
		job.put()

	logging.info('updating last_end to %s' % last_end)

	query = db.GqlQuery('SELECT * FROM Job WHERE crc = :1', crc)
	ent = query.get()
	db.run_in_transaction(tx, ent.key(), timestamp, last_end)

def add_suffix(crc, suffix):
	# races don't matter: when the suffix is set, nothing else matters
	job = get_job(crc)
	job.suffix = suffix
	job.put()

# TASK FUNCTIONS

def add_task(crc, mod_crc, begin, end):
	'''Add to the Task table. 'begin' and 'end' should be
	base64-encoded.'''

	ent = Task(crc=crc, mod_crc=mod_crc, begin=begin, end=end)
	ent.put()

def pending_tasks(crc):
	query = db.GqlQuery("SELECT * FROM Task WHERE crc = :1", crc)
	return query.fetch(100000000) #fetch needs a limit

# how many tasks going for this CRC
def count_pending_tasks(crc):
	query = db.GqlQuery('SELECT __key__ FROM Task WHERE crc = :1', crc)
	return query.count()

def remove_task(crc, end):
	'''Remove from the Task table. 'end' should be base64-encoded.'''

	query = db.GqlQuery('''
	    SELECT __key__ FROM Task
	    WHERE crc = :1 AND end = :2''', crc, end)
	ents = query.fetch(1000)
	#logging.info('I got %d results for (crc,end)=(%d,%s)' % (
	#    len(ents), crc, end))
	db.delete(ents)

# remove all tasks with this CRC
def clear_tasks(crc):
	query = db.GqlQuery('SELECT __key__ FROM Task WHERE crc = :1', crc)
	db.delete(query.fetch(1000))

def update_task_progress(crc, end, new_beginning):
	def tx(key):
		task = db.get(key)
		job.begin = new_beginning
		job.put()

	query = db.GqlQuery('''
	    SELECT * FROM Job
	    WHERE crc = :1 AND end = :2''',
	    crc, end)
	ent = query.get()
	db.run_in_transaction(tx, ent.key())
