# -*- coding: utf-8 -*-

# Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
# Lena Olson <lena@cs.wisc.edu>
#
# You can't do anything with this but read and execute it privately.  If you
# want to do more, ask us if we'd be willing to make a more permissive
# license.

import array
import base64
import cgi
import logging
import re
import time
import sys
import urllib

from google.appengine.api.labs import taskqueue
from google.appengine.ext import blobstore
from google.appengine.ext import webapp
from google.appengine.ext.webapp.util import run_wsgi_app
from google.appengine.runtime import DeadlineExceededError, apiproxy_errors

import datastore
import util

# check the cache,database every M,N iterations, where M,N are coprime
CACHE_CHECK_ITER = 7
DATABASE_CHECK_ITER = 16

MAX_TASK_QUEUE = 50

# how many values each task should be responsible for
TASK_SIZE = 1024 * 1024

# same as in collisionTryer.py
ASCII_MIN = 0x20
ASCII_MAX = 0x7f # non-inclusive
ASCII_RANGE = ASCII_MAX - ASCII_MIN # number of possible values

def xml_raw_response(completed, decode=False, crc=None, suffix=None,
    filename=None):
	if completed:
		comp_str = 'completed'
	else:
		comp_str = 'unfinished'
	res = "<?xml version='1.0' encoding='UTF-8'?>\n<check>\n"
	res += '\t<status>%s</status>\n' % comp_str
	if completed:
		# okay to decode suffix since it's printable now
		if decode:
			suffix = base64.b64decode(suffix)
		res += '\t<suffix>%s</suffix>\n' % urllib.quote(suffix)
		res += '\t<filename>%s</filename>\n' % urllib.quote(filename)
		download = '/download?crc=0x%08x' % crc
		res += '\t<download>%s</download>\n' % urllib.quote(download)
	res += '</check>\n'
	return res

def xml_response(job):
	if job.suffix is None: return xml_raw_response(False)
	#suffix = base64.b64decode(job.suffix)
	suffix = job.suffix
	decode = job.append_mode == 'ascii'

	return xml_raw_response(True, decode, job.crc, suffix,
	    util.filename(job))

def incr_job_binary(last, size):
	next = array.array('B', last)
	pos = len(next) - 1
	while size and pos >= 0:
		sum = size + next[pos]
		next[pos] = sum & 0xff
		pos -= 1
		size = sum >> 8
	while size:
		next.insert(0, size & 0xff)
		size >>= 8
	return next

def incr_job_ascii(last, size):
	next = array.array('B', last)
	pos = len(next) - 1
	while size and pos >= 0:
		sum = size + next[pos] - ASCII_MIN
		next[pos] = sum % ASCII_RANGE + ASCII_MIN
		pos -= 1
		size = sum / ASCII_RANGE
	while size:
		next.insert(0, size % ASCII_RANGE + ASCII_MIN)
		size /= ASCII_RANGE
	return next

def task_loop(crc, mod_crc, last_end, out_end, append_mode):
	pending_tasks = datastore.pending_tasks(crc)
	num_tasks = 0

	if append_mode == 'ascii':
		incr_job = incr_job_ascii
	elif append_mode == 'binary':
		incr_job = incr_job_binary
	else: assert 0

	# restart/resume dead tasks
	for task in pending_tasks:
		(running, saved_begin) = util.get_task_progress(crc, task.end)
		if not running:
			if saved_begin is None: begin = task.begin
			else: begin = saved_begin
			logging.info('restarting with begin=%s (was %s)' %
			    (saved_begin, task.begin))
			taskqueue.add(url='/tryToCollide', params={
			    'crc': crc, 'mod_crc': mod_crc,
			    'begin': begin, 'end': task.end,
			    'append_mode': append_mode,
			})
			# set to running in memcache so it doesn't get
			# restarted
			util.set_task_progress(crc, task.end, begin, True)
		num_tasks += 1

	# =0 so will check database and cache the first time
	i = 0

	begin = array.array('B', last_end)
	while True:
		if i % CACHE_CHECK_ITER == 0 and util.check_cache(crc):
			logging.info('found collision with memcache')
			break
		if i % DATABASE_CHECK_ITER == 0 and (
		    datastore.job_complete(crc)):
			logging.info('found collision with datastore')
			break
		if num_tasks == MAX_TASK_QUEUE:
			# wait a bit (via datastore query) and only return
			# if nothing could be done yet (which should never
			# happen)
			num_tasks0 = datastore.count_pending_tasks(crc)
			if num_tasks == num_tasks0: break

		# set end to the next begin + TASK_SIZE
		end = incr_job(begin, TASK_SIZE)
		end_base64 = util.encode_array(end)
		datastore.add_task(crc, mod_crc, util.encode_array(begin),
		    end_base64)
		num_tasks += 1

		# update the value in the Master object; a bit of a race which
		# I ignore (exception thrown right where this comment is)
		out_end[0] = end

		# start task
		begin_base64 = util.encode_array(begin)
		taskqueue.add(url='/tryToCollide', params={
		    'crc': crc, 'mod_crc': mod_crc,
		    'begin': begin_base64, 'end': end_base64,
		    'append_mode': append_mode,
		})
		# set to running in memcache so it doesn't get restarted
		util.set_task_progress(crc, end_base64, begin_base64, True)

		# wait for number of tasks in queue to drop below
		# MAX_TASK_QUEUE; I don't know how it could be greater, but
		# just in case...
		#while num_tasks >= MAX_TASK_QUEUE:
		#	time.sleep(1)
		#	num_tasks = datastore.count_pending_tasks(crc)

		# move end value to begin, increment counter
		begin = end
		i += 1

class Master(webapp.RequestHandler):
	def get(self):
		try:
			crc = int(self.request.get('crc'), 0)
		except ValueError:
			self.response.set_status(400, 'bad CRC')
			return

		job = datastore.get_job(crc)
		if job is None:
			self.response.set_status(400, 'no such job')
			return

		self.response.headers['Content-Type'] = 'application/xml'
		if job.suffix:
			self.response.out.write(xml_response(job))
			return

		# not the first time Master has run, so update timestamp and
		# the *real* 'last_end'
		last_end = util.get_job_progress(crc)
		logging.info('get_job_progress returned %s' % last_end)
		if last_end is not None:
			datastore.update_timestamp(crc, int(time.time()),
			    last_end)
		else:
			last_end = job.last_end

		try:
			end = util.decode_array(last_end)
			out_end = [end]
			task_loop(crc, job.mod_crc, end, out_end,
			    job.append_mode)

			# update job
			logging.info('set_job_progress with %s' %
			    util.encode_array(out_end[0]))
			util.set_job_progress(crc,
			    util.encode_array(out_end[0]))

			self.response.out.write(xml_response(job))
			return
			
			#The suffix should be in the datastore if the memcache indicates done
			## XXX for debugging
			#job.suffix = 'XXX'
			#job.put()

		except DeadlineExceededError: pass
		except apiproxy_errors.DeadlineExceededError: pass

		# exception was thrown
		# XXX this stuff will never meet the cleanup deadline; abandon
		# XXX all hope, ye who enter

		# clean up
		logging.info('deadline exceeded, setting progress to %s' %
		    util.encode_array(end))
		util.set_job_progress(crc, util.encode_array(out_end[0]))

		# return success or failure
		self.response.out.write(xml_response(job))

application = webapp.WSGIApplication([('/check', Master)], debug=True)

def main():
	run_wsgi_app(application)

if __name__ == '__main__':
	main()
